#!/usr/bin/env python3

'''
Batch sign files with GPG.
'''

import argparse
import getpass
import logging
import os
import re
import subprocess
import sys


LOGGER = logging.getLogger(__name__)

GPG_BIN = '/usr/bin/gpg'
SIG_EXT = '.sig'

parser = argparse.ArgumentParser(
    description='Batch sign files with gpg.'
)
parser.add_argument(
    'filepaths', metavar='<filepath>', nargs='+',
    help='The files to sign.'
)
parser.add_argument(
    '-u', '--uid', metavar='<GPG user ID>', required=True,
    help='The GPG user ID of the signing key.'
)
parser.add_argument(
    '--no-verify', action='store_true',
    help=('''
        Skip GPG verfication of existing signatures. Only check timestamps. This
        is faster but it will not detect existing invalid signatures.
    ''')
)
parser.add_argument(
    '--no-agent', action='store_true',
    help=('Prompt for password instead of expecting the GPG agent to handle it.')
)


def get_gpg_output(*args):
    '''
    Run a GPG command and return it's output. This sets the locale to 'C' before
    running the command.
    '''
    cmd = [GPG_BIN, *args]
    gpg_env = dict(os.environ)
    gpg_env['LC_ALL'] = 'C'
    return subprocess.check_output(cmd, stderr=subprocess.STDOUT, env=gpg_env).decode('utf-8')


def get_fingerprints():
    '''
    Get a list of all key fingerprints to convert short and long IDs to their
    fingerprints.
    '''
    output = get_gpg_output('--list-keys')
    pat = re.compile(r'Key fingerprint = ([0-9A-F ]+)')
    for line in output.splitlines():
        match = pat.search(line)
        if match:
            yield match.group(1).replace(' ', '')


def get_keyid(path):
    '''
    Return key ids by parsing output of "gpg --list-packets <path>"
    '''
    output = get_gpg_output('--list-packets', path)
    pat = re.compile(r'keyid ([0-9A-F]+)')
    match = pat.search(output)
    if match:
        return match.group(1)
    return None


def batch_sign(uid, paths, passphrase=None, only_check_timestamps=False):
    '''
    Batch sign several files with the key matching the given UID.  If no
    passphrase is given then the user is prompted for one.  The passphrase is
    returned to avoid further prompts. If only_check_timestamps is True, then
    existing signatures will only be checked by comparing their timestamps to
    their target files. If the target file is newer, they will be recreated.
    '''

    # For comparing key IDs of signatures.
    compact_uid = uid.replace(' ', '')
    uid_len = len(compact_uid)

    for path in paths:
        # Verify existing signatures. This fails if the sig is absent, invalid or
        # not from the given uid. In each case a new signature will be created.
        sigpath = path + SIG_EXT

        if only_check_timestamps:
            if os.path.exists(sigpath) and os.path.getmtime(sigpath) >= os.path.getmtime(path):
                continue
        else:
            cmd = [GPG_BIN, '--verify', sigpath]
            try:
                subprocess.check_call(
                    cmd,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL
                )
            except subprocess.CalledProcessError:
                pass
            else:
                kid = get_keyid(sigpath)
                min_len = min(uid_len, len(kid))
                if min_len >= 8 and compact_uid[-min_len:] == kid[-min_len:]:
                    continue
#     with open(os.devnull, 'w') as f:
#       p = subprocess.Popen(cmd, stdout=f, stderr=f)
#       e = p.wait()
#       if e == 0:
#         continue

        sigpath = path + '.sig'
        try:
            os.remove(sigpath)
        except FileNotFoundError:
            pass

        LOGGER.info('Signing %s', path)
        # Agent password management.
        if passphrase is False:
            cmd = [GPG_BIN, '-sbu', uid, '--batch', path]
            subprocess.run(cmd, check=True)
            return passphrase

        # Manual password passing.
        if passphrase is None:
            passphrase = getpass.getpass(f'Enter passphrase for {uid}: ').encode('utf-8')
        elif not isinstance(passphrase, bytes):
            passphrase = bytes(passphrase, 'UTF-8')

        cmd = [
            GPG_BIN, '-sbu', uid, '--batch',
            '--pinentry-mode', 'loopback',
            '--passphrase-fd', '0',
            path
        ]
        with subprocess.Popen(cmd, stdin=subprocess.PIPE) as proc:
            proc.communicate(passphrase)
            ecode = proc.wait()
        if ecode != 0:
            raise RuntimeError(f'error: {GPG_BIN} exited with non-zero status ({ecode:d})')

    return passphrase


def main(args=None):
    '''
    Main.
    '''
    pargs = parser.parse_args(args)
    passphrase = None if pargs.no_agent else False
    batch_sign(
        pargs.uid,
        pargs.filepaths,
        passphrase=passphrase,
        only_check_timestamps=pargs.no_verify
    )


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        pass
